#include <stdio.h>
#include <omp.h>

#define N 1024

double OUTopt[N][N], IN[N][N], W[N][N], OUTbase[N][N];

void init()
{
  for (int i=0; i<N; i++)
    for (int k=0; k<N; k++) {
	IN[i][k] = (double)i+1.0; 
	W[i][k] = (double)i+1.0; 
    }
}

int cmp (int k)
{
  int equal=1;
  for (int i=k; (i<N-k && equal); i++) {
    for (int j=k; (j<N-k && equal); k++) {
	if (OUTbase[i][j] != OUTopt[i][j])
		equal=0;
    }
  }
  return equal;
}

void base(int k)
{
  #pragma omp parallel
  #pragma omp single
  #pragma omp taskloop
  for (int i=k; i<N-k; i++)
    for (int j=k; j<N-k; j++) {
       OUTbase[i][j] = 0;
       for (int ii=-k; ii<=k; ii++)
         for (int jj=-k; jj<=k; jj++)
             OUTbase[i][j] += IN[i+ii][j+jj]*W[k+ii][k+jj];
    }
}

void STMT(int lb, int ub, int k, int i, int j) 
{
   for (int ii=-k; ii<=k; ii++) {
     for (int jj=lb; jj<=ub; jj++) {
       OUTopt[i][j-jj] += IN[i+ii][j]*W[k+ii][k+jj];
     }
   }
}

void opt (int k)
{
#pragma omp parallel
#pragma omp single
#pragma omp taskloop
  for (int i=k; i<N-k; i++) {
    for (int j=0; j<2*k; j++) { 
       OUTopt[i][j+k] = 0; 
       STMT(-k, -k+j, k, i, j); 
    }
    for (int j=2*k; j<N-2*k; j++) { 
       OUTopt[i][j+k] = 0; 
       STMT(-k, k, k, i, j); 
    }
    for (int j=N-2*k; j<N; j++)
       STMT(j-N+k+1, k, k, i, j);
  }
}

int main(int argc, char *argv[])
{
	double ini, end, tbase, topt;
	int k = 3;
	
	if (argc>1) k = atoi(argv[1]);

 	printf("\nSetting k=%d ", k);

        // cache warm-up	
	init();

	// execution with basic code
	ini=omp_get_wtime();
	base (k);
	end=omp_get_wtime();
	tbase = end-ini;

	// execution with optimized code
	ini=omp_get_wtime();
	opt (k);
	end=omp_get_wtime();
	topt = end-ini;

	// check reproducibility
	int res = cmp (k);

	int nths = omp_get_max_threads();

	if (res>0) {
	   printf("\nOK! %d %d %.8lf %.8lf\n", k, nths, tbase, topt);
           printf("\nTime %.8lf %.8lf\n", tbase, topt);
	}
	else 
           printf("\nNoOK! %d %d %.8lf %.8lf\n", k, nths, tbase, topt);
}

