#ifndef CONT_MATRIX_HH
#define CONT_MATRIX_HH

#include <vector>
#include <stdexcept>

#include <cstdlib>

template <typename T>
class ContMatrix {

public:
    ContMatrix(size_t rows, size_t cols): rows_(rows), cols_(cols) {
        data_ = static_cast<T*>(std::malloc(rows * cols * sizeof(T)));
        if (!data_) {
            throw std::bad_alloc();
        }
    }

    ~ContMatrix() {
        std::free(data_);
    }

    T& operator()(size_t row, size_t col) {
        if (row >= rows_ || col >= cols_) {
            throw std::out_of_range("Matrix indices out of range");
        }
        return data_[row * cols_ + col];
    }

    const T& operator()(size_t row, size_t col) const {
        if (row >= rows_ || col >= cols_) {
            throw std::out_of_range("Matrix indices out of range");
        }
        return data_[row * cols_ + col];
    }

    size_t rows() const { return rows_; }
    size_t cols() const { return cols_; }

private:
    size_t rows_;
    size_t cols_;
    T* data_;
};

#endif // CONT_MATRIX_HH