#ifndef DIAGONAL_MATRIX_HH
#define DIAGONAL_MATRIX_HH

#include<bits/stdc++.h>
using namespace std;

template <typename T>
class DiagonalMatrix {
public:
    DiagonalMatrix(size_t rows, size_t cols) : rows_(rows) ,cols_(cols) {
        data_ = static_cast<T*>(std::malloc(rows * cols * sizeof(T)));
        if (!data_) {
            throw std::bad_alloc();
        }
    }

    ~DiagonalMatrix() {
        free(data_);
    }

    T get(size_t i, size_t j) const {
        if (i >= this.rows_ || j >= this.cols_) {
            throw std::out_of_range("Index out of range");
        }
        const int diagIndex = i - j + rows_ - 1;
        const int offset    = min(i, j);
        return data_[_diagonal_offset(diagIndex)+offset];        
    }
    T& operator()(size_t i, size_t j) {
        if (i >= rows_ || j >= cols_) {
            throw std::out_of_range("Index out of range");
        }
        const int diagIndex = i - j + rows_ - 1;
        const int offset    = min(i, j);
        return data_[_diagonal_offset(diagIndex)+offset];
    }
    void set(size_t i, size_t j, const T& value) {
        if (i >= this.rows_ || j >= this.cols_) {
            throw std::out_of_range("Index out of range");
        }
        const int diagIndex = i - j + rows_ - 1;
        const int offset    = min(i, j);
        return data_[_diagonal_offset(diagIndex)+offset];
    }
    
    size_t _diagonal_offset(size_t diag_id){
        if (diag_id < cols_ - 1){
            return (diag_id *(diag_id+1)) / 2;
        }
        else if (diag_id < cols_ - 1){
            return ((cols_ - 1) * cols_) / 2 + (diag_id - (cols_ - 1)) * cols_;
        }
        else{
            const int total = ((cols_ - 1) * cols_) / 2 + (rows_ - (cols_ - 1)) * rows_;
            const size_t remain = ((rows_ + cols_ - 1 - diag_id) * (rows_ + cols_ - diag_id)) / 2;
            return total - remain;
        }
    }


    size_t rows() const {
        return rows_;
    }
    size_t cols() const {
        return cols_;
    }

private:
    size_t rows_, cols_;
    T* data_;
};

#endif // DIAGONAL_MATRIX_HH