#ifndef HALO_MATRIX2_HH
#define HALO_MATRIX2_HH

#include <stdexcept>
#include <cstdlib>

using namespace std;

#include <vector>

template <typename T>
class HaloMatrix2 {

public:
    HaloMatrix2(size_t rows, size_t cols): rows_(rows), cols_(cols), 
        body_((rows - 2) * (cols - 2)), 
        north_(rows), 
        south_(rows), 
        east_(cols), 
        west_(cols) {}

    ~HaloMatrix2() {

    }

    T& operator()(size_t i, size_t j) {
        if (i >= rows_ || j >= cols_) {
            throw std::out_of_range("Matrix indices out of range");
        }
        if (i > 0 && i < rows_ - 1 && j > 0 && j < cols_ - 1) {
            return body_[(i - 1) * (cols_ - 2) + (j - 1)];
        }
        return (i == 0)     ? north_[j]  :
            (i == rows_-1) ? south_[j]  :
            (j == 0)     ? west_[i]   :
                            east_[i];
    }

    const T& operator()(size_t i, size_t j) const {
        if (i >= rows_ || j >= cols_) {
            throw std::out_of_range("Matrix indices out of range");
        }
        if (i > 0 && i < rows_ - 1 && j > 0 && j < cols_ - 1) {
            return body_[(i - 1) * (cols_ - 2) + (j - 1)];
        }
        return (i == 0)     ? north_[j]  :
            (i == rows_-1) ? south_[j]  :
            (j == 0)     ? west_[i]   :
                            east_[i];
    }

    size_t rows() const { return rows_; }
    size_t cols() const { return cols_; }
    std::vector<T>& get_north() { return north_; }
    std::vector<T>& get_south() { return south_; }
    std::vector<T>& get_east()  { 
        east_[0] = north_[cols_ - 1];
        east_[rows_-1] = south_[cols_ - 1];
        return east_; 
    }
    std::vector<T>& get_west()  { 
        west_[0] = north_[0];
        west_[rows_ - 1] = south_[0];
        return west_; 
    }
    T& get_body(int i, int j) { return body_[(i - 1) * (cols_ - 2) + (j - 1)]; }


private:
    size_t rows_;
    size_t cols_;
    std::vector<T> body_;
    std::vector<T> north_;
    std::vector<T> south_;
    std::vector<T> east_;
    std::vector<T> west_;
};

#endif // HALO_MATRIX_HH