#ifndef INDIR_MATRIX_HH
#define INDIR_MATRIX_HH

#include <vector>
#include <stdexcept>

#include <cstdlib>

template <typename T>
class IndirectionMatrix {

public:
    IndirectionMatrix(size_t rows, size_t cols): rows_(rows), cols_(cols) {
        data_ = static_cast<T**>(std::malloc(rows * sizeof(T*)));
        if (!data_) {
            throw std::bad_alloc();
        }
        for (size_t i = 0; i < rows; ++i) {
            data_[i] = static_cast<T*>(std::malloc(cols * sizeof(T)));
            if (!data_[i]) {
                throw std::bad_alloc();
            }
        }
    }

    ~IndirectionMatrix() {
        for (size_t i = 0; i < rows_; ++i) {
            std::free(data_[i]);
        }
        std::free(data_);
    }

    T& operator()(size_t row, size_t col) {
        if (row >= rows_ || col >= cols_) {
            throw std::out_of_range("Matrix indices out of range");
        }
        return data_[row][col];
    }

    const T& operator()(size_t row, size_t col) const {
        if (row >= rows_ || col >= cols_) {
            throw std::out_of_range("Matrix indices out of range");
        }
        return data_[row][col];
    }

    size_t rows() const { return rows_; }
    size_t cols() const { return cols_; }

private:
    size_t rows_;
    size_t cols_;
    T** data_;
};

#endif