#include <bits/stdc++.h>
#include "cont_matrix.hh"
#include "indirection_matrix.hh"
#include "diagonal_matrix.hh"
#include "halo_matrix2.hh"

#ifdef __EXTRAE__
    #include <extrae.h>
#endif

#ifdef __DP__
    #define WP double
#else
    #define WP float
#endif

using namespace std;

int main() {
    const int rows = std::getenv("MAT_ROWS") ? std::stoi(std::getenv("MAT_ROWS")) : 10;
    const int cols = std::getenv("MAT_COLS") ? std::stoi(std::getenv("MAT_COLS")) : 10;
    
    #ifdef __EXTRAE__
    Extrae_eventandcounters(1000, 1);
    #else
    auto start = chrono::high_resolution_clock::now();
    #endif
    
    cout << "Using matrix" << rows << "x" << cols << endl;
    cout << "Using precision with " << sizeof(WP) << "B" << endl;

    #ifdef __indir_matrix
    IndirectionMatrix<WP> matA(rows, cols);
    IndirectionMatrix<WP> matB(rows, cols);
    cout << "Using indirect matrix" << endl;
    #elif defined(__halo_body)
    HaloMatrix2<WP> matA(rows, cols);
    HaloMatrix2<WP> matB(rows, cols);
    cout << "Using Halo matrix" << endl;
    #elif defined(__diag_matrix)
    DiagonalMatrix<WP> matA(rows, cols);
    DiagonalMatrix<WP> matB(rows, cols);
    cout << "Using diagonal matrix" << endl;
    #else
    ContMatrix<WP> matA(rows, cols);
    ContMatrix<WP> matB(rows, cols);
    cout << "Using contiguous matrix" << endl;
    #endif

    #ifdef __EXTRAE__
    Extrae_eventandcounters(1000, 0);
    Extrae_eventandcounters(4000, 1);
    #else
    auto end = std::chrono::high_resolution_clock::now();
    std::chrono::duration<double> elapsed = end - start;
    cout << "Allocate: " << elapsed.count() << " s" << endl;
    start = chrono::high_resolution_clock::now();
    #endif

    
    srand(33);
    #ifdef __halo_body
    for(int j = 0; j < cols; ++j){
        matA.get_north()[j] = static_cast<WP>(rand()) / static_cast<WP>(RAND_MAX);
        matA.get_south()[j] = static_cast<WP>(rand()) / static_cast<WP>(RAND_MAX);
        matB.get_north()[j] = static_cast<WP>(rand()) / static_cast<WP>(RAND_MAX);
        matB.get_south()[j] = static_cast<WP>(rand()) / static_cast<WP>(RAND_MAX);
    }
    for(int i = 1; i < rows - 1; ++i){
        matA.get_west()[i] = static_cast<WP>(rand()) / static_cast<WP>(RAND_MAX);
        matA.get_east()[i] = static_cast<WP>(rand()) / static_cast<WP>(RAND_MAX);
        matB.get_west()[i] = static_cast<WP>(rand()) / static_cast<WP>(RAND_MAX);
        matB.get_east()[i] = static_cast<WP>(rand()) / static_cast<WP>(RAND_MAX);
    }
    for (int i = 1; i < rows - 1; ++i) {
        for (int j = 1; j < cols - 1; ++j) {
            matA.get_body(i, j) = static_cast<WP>(rand()) / static_cast<WP>(RAND_MAX);
            matB.get_body(i, j) = static_cast<WP>(rand()) / static_cast<WP>(RAND_MAX);
        }
    }
    #else
    for (int i = 0; i < rows; ++i) {
        for (int j = 0; j < cols; ++j) {
            matA(i, j) = static_cast<WP>(rand()) / static_cast<WP>(RAND_MAX);
            matB(i, j) = static_cast<WP>(rand()) / static_cast<WP>(RAND_MAX);
        }
    }
    #endif

    ContMatrix<WP> result(rows, cols);

    #ifdef __EXTRAE__
    Extrae_eventandcounters(4000, 0);
    Extrae_eventandcounters(33, 1);
    #else
    end = std::chrono::high_resolution_clock::now();
    elapsed = end - start;
    cout << "Init: " << elapsed.count() << " s" << endl;
    start = chrono::high_resolution_clock::now();
    #endif

    for (int i = 0; i < rows - 1; ++i) {
        for (int j = 0; j < cols - 1; ++j) {
            result(i, j) =  matA(i, j) + matA(i + 1, j) + matA(i, j + 1) +
                            matB(i,j+1) + matB(i+1,j);
        }
    }
    #ifdef __EXTRAE__
    Extrae_eventandcounters(33, 0);
    Extrae_eventandcounters(22, 1);
    #else
    end = std::chrono::high_resolution_clock::now();
    elapsed = end - start;
    cout << "Compute stencil: " << elapsed.count() << " s" << endl;
    start = chrono::high_resolution_clock::now();
    #endif

    WP sum = 0;
    for (int i = 0; i < rows; ++i) {
        for (int j = 0; j < cols; ++j) {
            sum += result(i, j);
        }
    }

    #ifdef __EXTRAE__
    Extrae_eventandcounters(22, 0);
    #else
    end = std::chrono::high_resolution_clock::now();
    elapsed = end - start;
    cout << "Compute reduction: " << elapsed.count() << " s" << endl;
    #endif

    cout << "Sum of all values in the result matrix: " << sum << endl;

    // Print the result matrix
    //for (int i = 0; i < rows; ++i) {
    //    for (int j = 0; j < cols; ++j) {
    //        cout << result(i, j) << " ";
    //    }
    //    cout << endl;
    //}
    #ifdef __EXTRAE__
    Extrae_fini();
    #endif

    return 0;
}